/* Copyright (c) 2009 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Barry Leslie
 *
 * System dump table.
 *
 */
#include "CSConfig.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>

//#include "mysql_priv.h"
#include "CSGlobal.h"
#include "CSStrUtil.h"

#include "ha_pbms.h"
//#include <plugin.h>

#include "ms_mysql.h"
#include "Repository_ms.h"
#include "Database_ms.h"
#include "Compactor_ms.h"
#include "OpenTable_ms.h"
#include "Util_ms.h"
#include "Discover_ms.h"
#include "Transaction_ms.h"


#include "SysTab_dump.h"


DT_FIELD_INFO pbms_dump_info[]=
{
	{"Data",			NULL, NULL, MYSQL_TYPE_LONG_BLOB,	&my_charset_bin,	NOT_NULL_FLAG,	"A BLOB repository record"},
	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
};

DT_KEY_INFO pbms_dump_keys[]=
{
	{NULL, 0, {NULL}}
};


/*
 * -------------------------------------------------------------------------
 * DUMP TABLE
 */
//-----------------------
MSDumpTable::MSDumpTable(MSSystemTableShare *share, TABLE *table):MSRepositoryTable(share, table)
{
}

//-----------------------
MSDumpTable::~MSDumpTable()
{	
}

//-----------------------
void MSDumpTable::use()
{	
	dt_hasInfo = false;
	dt_headerSize = 0;
	
	// Suspend the transaction writer while the dump is running.
	MSTransactionManager::suspend(true);

	MSRepositoryTable::use();
}

//-----------------------
void MSDumpTable::unuse()
{
	MSTransactionManager::resume();
	MSRepositoryTable::unuse();
}

//-----------------------
void MSDumpTable::seqScanInit()
{
	dt_hasInfo = false;
	return MSRepositoryTable::seqScanInit();
}
//-----------------------
bool MSDumpTable::seqScanNext(char *buf)
{
	if (!dt_hasInfo) {
		dt_hasInfo = true;
		return returnInfoRow(buf);
	}
	// Reset the position
	return MSRepositoryTable::seqScanNext(buf);
}

//-----------------------
bool MSDumpTable::returnDumpRow(char *record, csWord8 record_size, char *buf)
{
	TABLE		*table = mySQLTable;
	Field		*curr_field;
	byte		*save;
	MY_BITMAP	*save_write_set;

	/* ASSERT_COLUMN_MARKED_FOR_WRITE is failing when
	 * I use store()!??
	 * But I want to use it! :(
	 */
	save_write_set = table->write_set;
	table->write_set = NULL;
	memset(buf, 0xFF, table->s->null_bytes);
	
 	for (Field **field=table->field ; *field ; field++) {
 		curr_field = *field;

		save = curr_field->ptr;
#if MYSQL_VERSION_ID < 50114
		curr_field->ptr = (byte *) buf + curr_field->offset();
#else
		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
#endif
		switch (curr_field->field_name[0]) {
			case 'D':
			case 'd':
				// Data         LONGBLOB
				ASSERT(strcmp(curr_field->field_name, "Data") == 0);
				if (record_size <= 0xFFFFFFF) {
					((Field_blob *) curr_field)->set_ptr(record_size, (byte *) record);
					ms_my_set_notnull_in_record(curr_field, buf);
				}
				break;
		}
		curr_field->ptr = save;
	}

	table->write_set = save_write_set;
	return true;
}

//-----------------------
bool MSDumpTable::returnRow(MSBlobHeadPtr blob, char *buf)
{
	csWord8		record_size;

	record_size = CS_GET_DISK_2(blob->rb_head_size_2) + CS_GET_DISK_6(blob->rb_blob_size_6);
	iBlobBuffer->setLength(record_size);
	iRepoFile->read(iBlobBuffer->getBuffer(0), iRepoOffset, (size_t) record_size, record_size);

	return returnDumpRow(iBlobBuffer->getBuffer(0), record_size, buf);
}

//-----------------------
#define INC_INFO_SPACE(i) record_size+=i; space-=i;ptr+=i;
#define MS_DUMP_MAGIC 0x5A74C1EB
typedef struct {
	CSDiskValue4 ti_table_id_4;
	char		 ti_name[1]; // variable length buffer
} TabInfoRec, *TabInfoPtr;

typedef struct {
	CSDiskValue4 di_magic_4;
	CSDiskValue2 di_header_size_2;
}	RepInfoRec, *RepInfoPtr;

// Repository DUMP info record format: <Dump magic><BLOB header size>[<table ID><table name>]...
bool MSDumpTable::returnInfoRow(char *buf)
{
	csWord8		record_size = 0, space = 1024;
	char		*ptr;
	MSTable		*tab;
	csWord4		space_needed, next_tab = 0;
	RepInfoPtr	rep_info;
	TabInfoPtr	tab_info;
	enter_();

	iBlobBuffer->setLength(space);
	ptr = iBlobBuffer->getBuffer(0);
	rep_info = (RepInfoPtr) iBlobBuffer->getBuffer(0);
	
	
	CS_SET_DISK_4(rep_info->di_magic_4, MS_DUMP_MAGIC);
	CS_SET_DISK_2(rep_info->di_header_size_2, sizeof(MSBlobHeadRec));
	
	INC_INFO_SPACE(sizeof(RepInfoRec));
	tab_info = (TabInfoPtr)ptr;
	
	while ((tab = myShare->mySysDatabase->getNextTable(&next_tab))) {
		push_(tab);
		space_needed = tab->myTableName->length() + 5;
		if (space < space_needed) {
			space += 1024;
			iBlobBuffer->setLength(space);
			ptr = iBlobBuffer->getBuffer(0) + record_size;
		}
		
		tab_info = (TabInfoPtr)ptr;
		CS_SET_DISK_4(tab_info->ti_table_id_4, tab->myTableID);
		strcpy(tab_info->ti_name, tab->myTableName->getCString());
		INC_INFO_SPACE(space_needed);
		
		release_(tab);
	}
	
	return_( returnDumpRow(iBlobBuffer->getBuffer(0), record_size, buf));
}

#define INC_INFO_REC(i) info_buffer+=i; length-=i; tab_info = (TabInfoPtr) info_buffer;
//-----------------------
void MSDumpTable::setUpRepository(const char *info_buffer, csWord4 length)
{
	csWord4 tab_id, magic;
	MSDatabase *myDB = myShare->mySysDatabase;
	RepInfoPtr	rep_info = (RepInfoPtr) info_buffer;
	TabInfoPtr	tab_info;
	
	if (length < sizeof(RepInfoRec)) {
		CSException::throwException(CS_CONTEXT, CS_ERR_INVALID_RECORD, "Invalid repository info record.");
	}

	magic = CS_GET_DISK_4(rep_info->di_magic_4);
	if (CS_GET_DISK_4(rep_info->di_magic_4) != MS_DUMP_MAGIC) {
		CSException::throwException(CS_CONTEXT, CS_ERR_BAD_HEADER_MAGIC, "Invalid repository info record.");
	}
	
	dt_headerSize = CS_GET_DISK_2(rep_info->di_header_size_2);
	INC_INFO_REC(sizeof(RepInfoRec));
	
	while (length > 5) {
		tab_id = CS_GET_DISK_4(tab_info->ti_table_id_4);
		myDB->addTable(tab_id, tab_info->ti_name, 0, false);
		INC_INFO_REC(strlen(tab_info->ti_name) +5);
	}
	
	if (length)
		CSException::throwException(CS_CONTEXT, CS_ERR_INVALID_RECORD, "Invalid repository info record.");		
}


//-----------------------
void MSDumpTable::insertRow(char *buf)
{	
	TABLE	*table = mySQLTable;
	Field_blob *field;
	csWord4 packlength, length;
	const char *blob_rec, *blob_ptr;
	
	field = (Field_blob *)(table->field[0]);
	
    /* Get the blob record: */
    blob_rec= buf + field->offset(table->record[0]);
    packlength= field->pack_length() - table->s->blob_ptr_size;

    memcpy(&blob_ptr, blob_rec +packlength, sizeof(char*));
    length= field->get_length();
	
	if (!dt_hasInfo) {
		setUpRepository(blob_ptr, length);
		dt_hasInfo = true;
	} else
		insertRepoRow((MSBlobHeadPtr)blob_ptr, length);
	
}

//-----------------------
void MSDumpTable::insertRepoRow(MSBlobHeadPtr blob, csWord4 length)
{	
	MSRepository *repo;
	MSRepoFile *repo_file;
	csWord8		repo_offset;
	csWord8		blob_size;
	csWord4		auth_code;
	csWord2 ref_size, ref_count, refs = 0, table_refs = 0;
	MSRepoPointersRec	ptr;
	MSDatabase *myDB = myShare->mySysDatabase;
	enter_();

	
	if (length != (CS_GET_DISK_2(blob->rb_head_size_2) + CS_GET_DISK_6(blob->rb_blob_size_6))) {
		CSException::throwException(CS_CONTEXT, MS_ERR_INVALID_RECORD, "Damaged Repository record");
	}
		
	// Get a repository file.
	repo = myDB->lockRepo(length);
	frompool_(repo);
	repo_file = repo->openRepoFile();
	push_(repo_file);

	repo_offset = repo->myRepoFileSize;
	
	// Reset the references for the BLOB and recreate
	// the temp log references.
	auth_code = CS_GET_DISK_4(blob->rb_auth_code_4);
	ref_count = CS_GET_DISK_2(blob->rb_ref_count_2);
	ref_size = CS_GET_DISK_1(blob->rb_ref_size_1);
	blob_size = CS_GET_DISK_6(blob->rb_blob_data_size_6);

	// First check to see if the BLOB is referenced
	ptr.rp_chars = ((char*) blob) + dt_headerSize;
	for (int count = 0; count < ref_count; count++) {
		int ref_type = CS_GET_DISK_2(ptr.rp_ref->rr_type_2);
		
		switch (ref_type) {
			case MS_BLOB_TABLE_REF:
				table_refs++;
				break;
				
			case MS_BLOB_FREE_REF:
			case MS_BLOB_DELETE_REF:
			case MS_BLOB_CLOUD_REF:
				break;
		
			default: // Assumed to be a MSRepoBlobRefRec.
				// Only committed references are backed up.
				if (IS_COMMITTED(CS_GET_DISK_8(ptr.rp_blob_ref->er_blob_ref_id_8))) {
					refs++;
				} else {
					CS_SET_DISK_2(ptr.rp_ref->rr_type_2, MS_BLOB_FREE_REF);
				}
				
				break;
		}
		
		ptr.rp_chars += ref_size;		
	}
	
	
	if (refs && table_refs) { // Unreferenced BLOBs are ignored.
	
	
		// Set table references.
		ptr.rp_chars = ((char*) blob) + dt_headerSize;
		for (int count = 0; count < ref_count; count++) {
			int ref_type = CS_GET_DISK_2(ptr.rp_ref->rr_type_2);
			MSOpenTable	*otab;
			csWord4		tab_id;
			csWord8		blob_id;
			csWord4		log_id; 
			csWord4		log_offset;
			csWord4		temp_time;
			csWord2		tab_index;
			
			
			
			switch (ref_type) {
				case MS_BLOB_TABLE_REF:
					tab_id = CS_GET_DISK_4(ptr.rp_tab_ref->tr_table_id_4);
					blob_id = CS_GET_DISK_6(ptr.rp_tab_ref->tr_blob_id_6);
					otab = MSTableList::getOpenTableByID(myDB->myDatabaseID, tab_id);
			
					frompool_(otab);
					otab->getDBTable()->setBlobHandle(otab, blob_id, repo->myRepoID, repo_offset, blob_size, dt_headerSize, auth_code);
					backtopool_(otab);
					break;
					
				case MS_BLOB_DELETE_REF:
				case MS_BLOB_CLOUD_REF:
					break;
			
			/*
				case MS_BLOB_CLOUD_REF:
				otab = MSTableList::getOpenTableForDB(myDB->myDatabaseID);
					frompool_(otab);
					myDB->queueCloudEvent(otab, repo->myRepoID, repo_offset, auth_code, &log_id, &log_offset, &temp_time);
					backtopool_(otab);
					
					repo->myLastTempTime = temp_time;
					CS_SET_DISK_4(ptr.rp_temp_ref->tp_log_id_4, log_id);
					CS_SET_DISK_4(ptr.rp_temp_ref->tp_offset_4, log_offset);
					break;
			*/
					
				case MS_BLOB_FREE_REF:
				default: 
					break;
			}
				
			ptr.rp_chars += ref_size;		
		}	
	
		// Write the repository record.
		repo_file->write(blob, repo_offset, length);
		repo->myRepoFileSize += length;
		
		csWord2 alias_offset;
		if (alias_offset = CS_GET_DISK_2(blob->rb_alias_offset_2)) { 
			myDB->registerBlobAlias(repo->myRepoID, repo_offset, ((char*)blob) + alias_offset);
		}
	}

	release_(repo_file);
	backtopool_(repo);
	exit_();
}



