/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase S3Daemon
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  Created by Barry Leslie on 10/2/08.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <ctype.h>



#include "keys_s3.h"

#ifndef PBMS_CLOUD_ENCRYPT_KEY
#define PBMS_CLOUD_ENCRYPT_KEY "PBMS_CLOUD"
#endif

#define MAX_INFO_LINE (MAX_NODE_SIZE + MAX_PUBLIC_KEY + MAX_PRIVATE_KEY + 32) 
#define MIN_INFO_LINE 5

/*-----------------------------------------------------------*/
static csWord4 hash(const char *str)
{
	csWord4 v = 0;
	while (*str) {
		v += (v +1) * *str;
		str++;
	}
	return v;
}

/*-----------------------------------------------------------*/
/*
 * This is just to prevent someone from being able to findout what the key
 * is just by reading the file.
 * The real security is up to the user to make sure the key file is secure.
 */
static void key_encrypt(CSFile *info_file, CloudKey *key)
{
	csWord4 i;
	csWord4 *ptr, anum, len=0;
	CloudKey target = *key;
	enter_();
	push_(info_file);
	

	ptr = (unsigned long *)&target;
	len = sizeof(CloudKey)/sizeof(*ptr);
	
	anum = hash(PBMS_CLOUD_ENCRYPT_KEY);
	
	/* Fill the target with rubish */
	for (i = 0; i < len; i++, ptr++) *ptr = rand();

	strcpy(target.node, key->node);
	strcpy(target.public_key, key->public_key);
	strcpy(target.private_key, key->private_key);
	ptr = (unsigned long *)&target;

	for (i = 0; i < len; i++, ptr++) *ptr = (*ptr + anum) ^ anum;

	info_file->setEOF(0);
	info_file->setEOF(0);
	
	info_file->write("#", 0, 1);
	info_file->write(&target, 1, sizeof(S3Keys));
	
	release_(info_file);
	exit_();
}

/*-----------------------------------------------------------*/
static void unencryptCloudInfo(char *encrypt_key, CloudKey *info)
{
	int i;
	unsigned long *ptr, anum, len=0;
	
	ptr = (unsigned long *)info;
	len = sizeof(CloudKey)/sizeof(*ptr);
	memcpy(ptr, encrypt_key+1, sizeof(CloudKey));
	anum = hash(PBMS_CLOUD_ENCRYPT_KEY);
	
	for (i = 0; i < len; i++, ptr++) *ptr = (*ptr ^ anum) - anum;
	
}

/*-----------------------------------------------------------*/
void loadCloudInfo(CSFile *info_file, CloudKey *info)
{
	int cnt;
	char line[MAX_INFO_LINE], *ptr, *privkey, *pubkey;
	const char  *err = NULL;
	enter_();
	push_(info_file);
	
	cnt = info_file->read(line, 0, MAX_INFO_LINE, MIN_INFO_LINE);
	line[cnt] = 0;
	
	if (*line == '#') {
		key_unencrypt(line, keys);
		goto done;
	}
	
	err = 	"Unexpected EOL";
	
	ptr = line;
	while (isspace(*ptr)) ptr++;	
	if (!*ptr) {
		goto error;
	}
	
	pubkey = ptr;
	while (*ptr && !isspace(*ptr)) ptr++;
	if (!*ptr) 
		goto error;
	*ptr = 0;
	ptr++;

	while (isspace(*ptr)) ptr++;	
	if (!*ptr) 
		goto error;
	
	privkey = ptr;
	while (*ptr && !isspace(*ptr)) ptr++;
	if (*ptr) {
		*ptr = 0;
		ptr++;
		while (isspace(*ptr)) ptr++;
		if (*ptr && (ptr - line) < cnt) {	
			err = 	"Unexpected chatrater(s) after private key";
			goto error;
		}
	}
	
	if (strlen(pubkey) >= MAX_PUBLIC_KEY || strlen(privkey) >= MAX_PRIVATE_KEY ) {
		err = "Maximum key length exceeded.";
		goto error;
	}
	
	strcpy(keys->public_key, pubkey);
	strcpy(keys->private_key, privkey);
	
	info_file->retain();
	key_encrypt(info_file, keys);
	goto done;

error:
	{
		char msg[1024];
		cs_strcpy(1024, msg, info_file->getPathString());
		cs_strcat(1024, msg, ": ERROR:  ");
		cs_strcat(1024, msg, err);
		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, msg);
	}
done:	
	release_(info_file);
	exit_();
	
}

